(ns puppetlabs.rbac.storage.authorization
  (:require [clojure.set :refer [union difference]]
            [slingshot.slingshot :refer [throw+]]
            [puppetlabs.rbac.ldap :as ldap]
            [puppetlabs.rbac.shiro-utils :refer [is-permitted? are-permitted?]]
            [puppetlabs.rbac.storage :as st]
            [puppetlabs.rbac.sync :as sn]))

(def base-perms
  {#'st/get-groups "user_groups:view"
   #'st/get-group "user_groups:view"
   #'sn/create-group! "user_groups:import"
   #'st/update-remote! "user_roles:edit_members"
   #'st/get-users "users:view"
   #'st/get-user "users:view"
   #'st/update-user!"users:edit"
   #'st/get-directory-settings "directory_service:edit"
   #'st/update-directory-settings! "directory_service:edit"
   #'ldap/test-connection "directory_service:edit"})

(defn- permission-exception [perm]
  (throw+ {:kind :puppetlabs.rbac/permission-denied
           :msg "You do not have permission to perform this action."
           :permissions perm})) ;; TODO use descriptions from actions table in msg

(defn make-perm [base-perm instance]
  (format "%s:%s" base-perm instance))

(defn single-instance-check
  "For use with storage functions that simply need a true/false
  answer to is-permitted? before proceeding. The instance id
  of the object that is returned by the storage function needs
  to be under the key :id."
  [subject storage-fn & args]
  (let [base-perm   (base-perms storage-fn)
        object      (apply storage-fn args)
        instance-id (:id object)
        perm        (make-perm base-perm instance-id)]
    (if (is-permitted? subject perm)
      object
      (permission-exception perm))))

(defn no-instance-check
  "For use with storage functions that have associated permissions
  with no relevant instances (typically 'create' actions)."
  [subject storage-fn & args]
  (let [base-perm (base-perms storage-fn)
        perm      (make-perm base-perm "*")]
    (if (is-permitted? subject perm)
      (apply storage-fn args)
      (permission-exception perm))))

(defn multiple-instances-check
  "For use with storage functions that return a collection of instances."
  [subject storage-fn & args]
  (let [base-perm (base-perms storage-fn)
        coll      (apply storage-fn args)]

    (if (is-permitted? subject (make-perm base-perm "*"))
      coll
      (let [perm-strs   (map (comp #(make-perm base-perm %) :id) coll)
            result      (are-permitted? subject perm-strs)
            filter-coll (->> (zipmap coll result)
                             (filter second))]
        (map first filter-coll)))))

(defn update-group-check
  "Only for the storage function that updates groups. Checks that the roles
  which are added or removed are all permitted for the current subject."
  [subject storage-fn & [db id {:keys [role_ids]} :as args]]
  (let [new-roles   (set role_ids)
        exist-roles (set (:role_ids (st/get-group db :id id)))
        symm-diff   (union (difference new-roles exist-roles) (difference exist-roles new-roles))
        base-perm   (base-perms storage-fn)
        perms       (map #(make-perm base-perm %) symm-diff)]

    (if (every? true? (are-permitted? subject perms))
      (apply storage-fn args)
      (permission-exception perms))))
